local utf8 = require 'utf8'

local function newGrid(opt)
    local style = opt.style
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local sx,sy = style.scale.normal.x,style.scale.normal.y
    local offx,offy = style.offset.normal.x,style.offset.normal.y
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local font = opt.font.normal.obj
    local color = style.fontColor.normal
    local xx,yy,ww,hh = x+offx+2,y+offy+2,w*sx-4,h*sy-4
    local fw,fh = font:getWidth("中"),font:getHeight("中")
    local cols = math.floor(ww/fw *2)
    local magrin_c = (ww- cols*fw/2)/2
    opt.grid = {
        x = xx + magrin_c,
        y = yy + hh/2 - fh/2,
        w = cols * fw/2,
        h = fh,
        cw = fw/2,
        ch = fh,
        cols = cols,
    }
end

local function fillText(opt)
    opt.aline = opt.aline or "right"
    opt.grid.cells = {}
    local grid = opt.grid
    local cells = opt.grid.cells
    local count = utf8.len(opt.text)
    local text = opt.text
    local a = opt.aline
    for i = 1,grid.cols do
        cells[i] = {}
    end
    a = opt.editing and "center" or a
    if a == "left" then
        local p = 1
        for i = 1,count do
            local cell = cells[p]
            local l,r = utf8.offset(text,i),utf8.offset(text,i+1) -1
            local char = opt.text:sub(l,r)
            local tx = r - l == 0 and 1 or 2
            if cell then
                cell.l = l
                cell.r = r
                cell.t = true
                cell.char = char
                cell.w = tx
            end
            p = p + tx
        end
    elseif a == "right" then
        local p = grid.cols
        for i = count,1,-1 do
            local l,r = utf8.offset(text,i),utf8.offset(text,i+1) -1
            local char = opt.text:sub(l,r)
            local tx = r - l == 0 and 1 or 2
            local cell = tx == 1 and cells[p] or cells[p-1]
            if cell then
                cell.l = l
                cell.r = r
                cell.t = true
                cell.char = char
                cell.w = tx
            end
            p = p - tx
        end
    elseif a == "center" then
        local total = 0
        for i = 1,count do
            local l,r = utf8.offset(text,i),utf8.offset(text,i+1)
            local w = r-l > 1 and 2 or 1
            total = total + w
        end
        local p = math.floor(grid.cols/2 - total/2)
        for i = 1,count do
            local cell = cells[p]
            local l,r = utf8.offset(text,i),utf8.offset(text,i+1) -1
            local char = opt.text:sub(l,r)
            local tx = r - l == 0 and 1 or 2
            if cell then
                cell.l = l
                cell.r = r
                cell.t = true
                cell.char = char
                cell.w = tx
            end
            p = p + tx
        end
    else --free
        opt.offset = opt.offset or math.floor(grid.cols/2 - count/2)
        local p = 1 + opt.offset
        for i = 1,count do
            local cell = cells[p]
            
            local l,r = utf8.offset(text,i),utf8.offset(text,i+1) -1
            local char = opt.text:sub(l,r)
            local tx = r - l == 0 and 1 or 2
            if cell then
                cell.l = l
                cell.r = r
                cell.t = true
                cell.char = char
                cell.w = tx
            end
            p = p + tx
        end
    end
end

local function newCursor(opt)
    opt.cursor = {
        pos = 1,
        mode = "insert",
        timer = 0,
        cd = 1
        }
end

local function addText(opt,text)
    local cells = opt.grid.cells
    local cell = opt.grid.cells[opt.cursor.pos]
    local target = cell.l
    if not cell.t then
        opt.text = opt.text..text
        fillText(opt)
        local cells = opt.grid.cells
        for i = #cells,1,-1 do
            local cell = cells[i]
            if cell.char then
                opt.cursor.pos = i + cell.w
                break
            end
        end
    else
        local l,r = string.sub(opt.text,1,cell.l-1),string.sub(opt.text,cell.l,-1)
        opt.text = l..text..r
        fillText(opt)
        local plus = text:len()
        local cells = opt.grid.cells
        for i = 1,#cells do
            local cell = cells[i]
            if cell.l == target + plus then
                opt.cursor.pos = i
                break
            end
        end
    end
    
end



local function adjustCursor(opt)
    local c = opt.cursor
    local g = opt.grid
    local cells = g.cells
    if not cells[c.pos].char then
        if cells[c.pos-1] then
            if cells[c.pos-1].w == 2 then
                c.pos = c.pos - 1
            elseif not cells[c.pos-1].t then
                local l,r
                for i = 1,#cells do
                    if cells[i].t and not l then l = i end
                    if cells[i].t then r = i+cells[i].w end
                end
                if c.pos<l then c.pos = l end
                if c.pos>r then c.pos = r end
            end
        else
            local l,r
            for i = 1,#cells do
                if cells[i].t and not l then l = i end
                if cells[i].t then r = i+cells[i].w end
            end
            if c.pos<l then c.pos = l end
            if c.pos>r then c.pos = r end
        end
    end
end

local function updateCursor(opt,mx,my)
    local c = opt.cursor
    local g = opt.grid
    local cells = g.cells
    c.pos = math.ceil((mx - g.x)/g.cw) 
    c.pos = c.pos>#cells and #cells or c.pos
    c.pos = c.pos<1 and 1 or c.pos
    adjustCursor(opt)
end

local function delText(opt,pos)
    local cells = opt.grid.cells
    local cursor_pos = opt.cursor.pos
    local cursor_l = cells[cursor_pos].l
    local cell = cells[cursor_pos - 1]
    if not cell.char then
        cell = cells[cursor_pos - 2]
    end
    if not cell.char then return end
    local del_w = cell.r - cell.l + 1
    local l,r = opt.text:sub(1,cell.l-1),opt.text:sub(cell.r+1,-1)
    opt.text = l..r
    opt.text = opt.text == "" and " " or opt.text
    fillText(opt)
    local cells = opt.grid.cells
    if cursor_l then
        for i = 1,#cells do
            if cells[i].l == cursor_l - del_w then
                opt.cursor.pos = i
                adjustCursor(opt)
            end
        end
    else
        local l,r
        for i = 1,#cells do
            if cells[i].t and not l then l = i end
            if cells[i].t then r = i+cells[i].w end
        end
        opt.cursor.pos = r + 1
        adjustCursor(opt)
    end
end


local function keyConf(opt,key)
    if key == "backspace" then
        delText(opt)
    elseif key == "right" then
        local w = opt.grid.cells[opt.cursor.pos] and opt.grid.cells[opt.cursor.pos].w or 1
        opt.cursor.pos = math.min(#opt.grid.cells,opt.cursor.pos+w)
        adjustCursor(opt)
    elseif key == "left" then
        opt.cursor.pos = math.max(1,opt.cursor.pos-1)
        adjustCursor(opt)
    elseif key == "return" then
        opt.editing = false
    end
end

local name = "text"
return function(core,opt)
    opt.id = opt.id or tostring(name.."_"..tostring(opt))
    opt.font = opt.font or core.style.newFont()
    opt.state = opt.state or "normal"
    ---interact
    opt.submit = nil
    opt.editing = false
    if opt.disabled then
        opt.state = "disabled"
    else
        core:updateState(opt)
    end
    opt.editing = core:grabFocus(opt) 
    opt.active = opt.editing
    if opt.editing and core.text then
        if opt.cursor.mode == "insert" then
            addText(opt,core.text)
        else
            replaceText(opt)
        end
        
    end
    
    if opt.editing and core.key then
        if core.key == "return" then
            core:releaseFocus()
            opt.submit = opt.text
        else
            keyConf(opt,core.key)
        end
    end
    
    
    ---draw
    core:applyLayout(opt)
    opt.draw = opt.draw or core.render[name]
    opt.style = opt.style or core.style
    
    if not opt.buff or opt.dirty then
        newGrid(opt)
        fillText(opt)
    end
    if not opt.cursor then
        newCursor(opt)
    end
    if core:isActive(opt) then
        updateCursor(opt,core.mouse_x,core.mouse_y)
    end


    if core:gotFocus(opt) then
        fillText(opt)
    end
    
    if core:hadFocus(opt) then
        fillText(opt)
    end
    
    core:registerDraw(opt.draw ,{opt.style,opt})
    core:updateBound(opt)
    return opt
end